/*
  Copyright 2019 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.auth

import com.kakao.sdk.auth.model.AccessTokenResponse
import com.kakao.sdk.auth.model.AgtResponse
import com.kakao.sdk.auth.model.OAuthToken
import com.kakao.sdk.auth.network.kauth
import com.kakao.sdk.network.ApiFactory
import com.kakao.sdk.common.KakaoJson
import com.kakao.sdk.common.KakaoSdk
import com.kakao.sdk.common.model.*
import retrofit2.Call
import retrofit2.Callback
import retrofit2.HttpException
import retrofit2.Response

/**
 * @author kevin.kang. Created on 2019-11-01..
 */
class AuthApiClient(
    private val authApi: AuthApi = ApiFactory.kauth.create(AuthApi::class.java),
    private val accessTokenRepo: AccessTokenRepo = AccessTokenRepo.instance
) {
    fun issueAccessToken(
        authCode: String,
        redirectUri: String = "kakao${KakaoSdk.applicationContextInfo.appKey}://oauth",
        clientId: String = KakaoSdk.applicationContextInfo.appKey,
        androidKeyHash: String = KakaoSdk.applicationContextInfo.signingKeyHash,
        callback: (response: AccessTokenResponse?, error: Throwable?) -> Unit
    ) {
        authApi.issueAccessToken(
            clientId = clientId,
            redirectUri = redirectUri,
            androidKeyHash = androidKeyHash,
            authCode = authCode
        ).enqueue(object: Callback<AccessTokenResponse> {
            override fun onFailure(call: Call<AccessTokenResponse>, t: Throwable) {
                callback(null, t)
            }

            override fun onResponse(
                call: Call<AccessTokenResponse>,
                response: Response<AccessTokenResponse>
            ) {
                if (response.isSuccessful) {
                    response.body()?.let {
                        accessTokenRepo.setToken(OAuthToken.fromResponse(it))
                        callback(it, null)
                        return
                    }
                    callback(null, ClientError(ClientErrorCause.Unknown, "No body"))
                } else {
                    callback(null, translateError(HttpException(response)))
                }
            }
        })
    }

    fun refreshAccessToken(
        refreshToken: String,
        clientId: String = KakaoSdk.applicationContextInfo.appKey,
        androidKeyHash: String = KakaoSdk.applicationContextInfo.signingKeyHash,
        callback: (response: AccessTokenResponse?, error: Throwable?) -> Unit) {
        authApi.issueAccessToken(
            clientId = clientId,
            redirectUri = null,
            androidKeyHash = androidKeyHash,
            refreshToken = refreshToken,
            grantType = Constants.REFRESH_TOKEN
        ).enqueue(object : Callback<AccessTokenResponse> {
            override fun onFailure(call: Call<AccessTokenResponse>, t: Throwable) {
                callback(null, t)
            }

            override fun onResponse(
                call: Call<AccessTokenResponse>,
                response: Response<AccessTokenResponse>
            ) {
                if (response.isSuccessful) {
                    response.body()?.let {
                        accessTokenRepo.setToken(OAuthToken.fromResponse(it))
                        callback(it, null)
                        return
                    }
                    callback(null, ClientError(ClientErrorCause.Unknown, "No body"))
                } else {
                    callback(null, translateError(HttpException(response)))
                }
            }

        })
    }

    fun agt(
        clientId: String = KakaoSdk.applicationContextInfo.appKey,
        callback: (response: AgtResponse?, error: Throwable?) -> Unit
    ) {
        accessTokenRepo.getToken().accessToken?.let {
            authApi.agt(clientId, it).enqueue(object: Callback<AgtResponse> {
                override fun onFailure(call: Call<AgtResponse>, t: Throwable) {
                    callback(null, t)
                }

                override fun onResponse(call: Call<AgtResponse>, response: Response<AgtResponse>) {
                    response.body()?.let {
                        callback(it, null)
                        return
                    }
                    try {   // 예외케이스로 서버에서 JSON이 아니라 html을 내려줄 때 fromJson이 던지는 예외처리  TODO 다른 곳들도 대응 필요
                        callback(null, translateError(HttpException(response)))
                    } catch (error: Throwable) {
                        callback(null, ClientError(ClientErrorCause.Unknown, error.toString()))
                    }
                }
            })
            return
        }
        callback(null, ClientError(ClientErrorCause.TokenNotFound, "Access token not found."))
    }

    companion object {
        fun translateError(t: Throwable): Throwable {
            if (t is HttpException) {
                val errorString = t.response()?.errorBody()?.string()
                val response =
                    KakaoJson.fromJson<AuthErrorResponse>(errorString!!, AuthErrorResponse::class.java)
                val cause = KakaoJson.fromJson(response.error, AuthErrorCause::class.java)
                    ?: AuthErrorCause.Unknown
                return OAuthError(t.code(), cause, response)
            }
            return t
        }

        val instance by lazy { AuthApiClient() }
    }
}