package com.kakao.sdk.auth.network

import android.util.Log
import com.kakao.sdk.auth.AccessTokenRepo
import com.kakao.sdk.auth.AuthCodeClient
import com.kakao.sdk.auth.network.AccessTokenInterceptor.Companion.requestWithAuthorization
import com.kakao.sdk.common.KakaoJson
import com.kakao.sdk.common.KakaoSdk
import com.kakao.sdk.common.model.ApiErrorResponse
import com.kakao.sdk.common.model.ApplicationContextInfo
import okhttp3.Interceptor
import okhttp3.Response
import java.util.concurrent.CountDownLatch

class DynamicAgreementInterceptor(
    private val tokenRepo: AccessTokenRepo = AccessTokenRepo.instance,
    private val contextInfo: ApplicationContextInfo = KakaoSdk.applicationContextInfo
) : Interceptor {

    override fun intercept(chain: Interceptor.Chain): Response {
        val response = chain.proceed(chain.request())

        if (response.code() == 403) {
            val error = response.body()?.string()?.let {
                KakaoJson.fromJson<ApiErrorResponse>(it, ApiErrorResponse::class.java)
            }

            if (error?.requiredScopes != null) {
                Log.e("#### DYNAMIC AGREEMENT", "${error.requiredScopes}")

                val latch = CountDownLatch(1)
                var newAgreementError: Throwable? = null
                AuthCodeClient.instance.newAgreements(contextInfo.context, error.requiredScopes!!) { error ->
                    newAgreementError = error
                    latch.countDown()
                }
                latch.await()

                if (newAgreementError != null) {
                    throw java.io.IOException("Kakao SDK Error", newAgreementError!!)
                }

                val accessToken = tokenRepo.getToken().accessToken
                if (accessToken != null) {
                    return chain.proceed(requestWithAuthorization(response.request(), accessToken))
                } else {
                    Log.e("#### DYNAMIC AGREEMENT", "토큰이 없음")
                }
            }
        }

        return response
    }
}